import pandas as pd
import numpy as np
import os

def apply_metric_groups(data_set_path):

    score_save_path=data_set_path.replace('.csv','_scores.csv')
    assert os.path.isfile(score_save_path),'You must run listing 5.3 or 7.5 to save metric scores first'
    score_data = pd.read_csv(score_save_path,index_col=[0,1])
    data_2group = score_data.drop('is_churn',axis=1)

    load_mat_path = data_set_path.replace('.csv', '_load_mat.csv')
    assert os.path.isfile(load_mat_path),'You must run listing 6.4 to save a loading matrix first'
    load_mat_df = pd.read_csv(load_mat_path, index_col=0)
    load_mat_ndarray = load_mat_df.to_numpy()

    # Make sure the data is in the same column order as the rows of the loading matrix
    ndarray_2group = data_2group[load_mat_df.index.values].to_numpy()
    grouped_ndarray = np.matmul(ndarray_2group, load_mat_ndarray)

    churn_data_grouped = pd.DataFrame(grouped_ndarray,columns=load_mat_df.columns.values, index=score_data.index)

    churn_data_grouped['is_churn'] = score_data['is_churn']

    save_path = data_set_path.replace('.csv', '_groupscore.csv')
    churn_data_grouped.to_csv(save_path,header=True)
    print('Saved grouped data  to ' + save_path)





from datetime import date, datetime, timedelta,time
from dateutil import relativedelta
from numpy import random
from random import randrange
import numpy as np



class Customer:
    id_counter=0
    MIN_AGE = 12.0
    MAX_AGE = 82.0
    AGE_RANGE = MAX_AGE - MIN_AGE
    AVG_AGE = (MAX_AGE + MIN_AGE) / 2.0
    date_multipliers = {}

    def __init__(self,behavior_rates,satisfaction=None,channel_name='NA',start_of_month=None,country=None):
        '''
        Creates a customer for simulation, given an ndarray of behavior rates, which are converted to daily.
        Each customer also has a unique integer id which will become the account_id in the database, and holds its
        own subscriptions and events.
        :param behavior_rates: ndarray of behavior rates, which are assumed to be PER MONTH
        '''
        self.id=Customer.id_counter # set the id to the current class variable
        Customer.id_counter+=1 # increment the class variable

        self.behave_per_month=behavior_rates
        self.behave_per_day = (1.0/30.0)*self.behave_per_month
        self.channel=channel_name

        if start_of_month:
            self.age=random.uniform(Customer.MIN_AGE,Customer.MAX_AGE)
            self.date_of_birth = start_of_month + relativedelta.relativedelta(years=-int(self.age),
                                                                              months=-int( (self.age % 1)*12 ),
                                                                              days=-random.uniform(1,30))
        else:
            self.date_of_birth=None

        self.country=country

        if satisfaction is None:
            age_contrib = 0.5* (Customer.AVG_AGE - self.age)/Customer.AGE_RANGE
            self.satisfaction_propensity = np.power(2.0, random.uniform(-1.5, 1.5) + age_contrib)
        else:
            self.satisfaction_propensity = satisfaction
        self.subscriptions=[]
        self.events=[]

    def generate_events(self,start_date,end_date):
        '''
        Generate a sequence of events at the customers daily rates.  Each count for an event on a day is droing from
        a poisson distribution with the customers average rate.  If the number is greater than zero, that number of events
        are created as tuples of time stamps and the event index (which is the database type id).  The time of the
        event is randomly set to anything on the 24 hour range.
        :param start_date: datetime.date for start of simulation
        :param end_date: datetime.date for end of simulation
        :return: The total count of each event, the list of all of the event tuples
        '''

        delta = end_date - start_date

        events=[]
        counts=[0]*len(self.behave_per_day)
        for i in range(delta.days):
            the_date = start_date + timedelta(days=i)
            if the_date in Customer.date_multipliers:
                multiplier = Customer.date_multipliers[the_date]
            else:
                if the_date.weekday() >= 4:
                    multiplier = random.uniform(1.00,1.2)
                else:
                    multiplier = random.uniform(0.825,1.025)
                Customer.date_multipliers[the_date]=multiplier
            for event_idx,rate in  enumerate(self.behave_per_day):
                new_count= int(round(multiplier*random.poisson(rate)))
                counts[event_idx] += new_count
                for n in range(0,new_count):
                    event_time=datetime.combine(the_date,time(randrange(24),randrange(60),randrange(60)))
                    new_event=(event_time,event_idx)
                    events.append(new_event )

        self.events.extend(events)

        return counts


import os
import pickle
import shap
import matplotlib.pyplot as plt
from listing_8_4_rescore_metrics import reload_churn_data

def shap_explain_xgb(data_set_path, plot_n=None):

    pickle_path = data_set_path.replace('.csv', '_xgb_model.pkl')
    assert os.path.isfile(pickle_path), 'You must run listing 9.6 to save an XGB regression model first'
    with open(pickle_path, 'rb') as fid:
        xgb_model = pickle.load(fid)

    current_df = reload_churn_data(data_set_path,'current','8.3',is_customer_data=True)
    explainer = shap.TreeExplainer(xgb_model,feature_perturbation = "interventional", model_output='probability',data=current_df)
    shap_values = explainer(current_df)
    shap.summary_plot(shap_values, current_df, show=False)
    save_file = data_set_path.replace('.csv', '_shap_summary_xgb.png')
    print(f'Saving SHAP Explanation to {save_file}')
    plt.tight_layout()
    plt.savefig(save_file, format='png')
    plt.close()

    if plot_n is not None:
        for n in plot_n:
            shap.waterfall_plot(shap_values[n],show=False)
            save_file = data_set_path.replace('.csv', f'_shap_water_xgb_{n}.png')
            plt.tight_layout()
            plt.savefig(save_file, format='png')
            plt.close()



import pandas as pd

from listing_10_3_grouped_category_cohorts import group_category_column

def dummy_variables(data_set_path, groups={},current=False):

    raw_data = pd.read_csv(data_set_path, index_col=[0, 1])

    for cat in groups.keys():
        group_category_column(raw_data,cat,groups[cat])

    data_w_dummies = pd.get_dummies(raw_data,dummy_na=True)
    new_cols = sorted(list(set(data_w_dummies.columns).difference(set(raw_data.columns))))
    cat_cols = sorted(list(set(raw_data.columns).difference(set(data_w_dummies.columns))))
    dummy_col_df = pd.DataFrame(new_cols,index=new_cols,columns=['metrics'])
    dummy_col_df.to_csv(data_set_path.replace('.csv', '_dummies_groupmets.csv'))

    if not current:
        new_cols.append('is_churn')
        data_w_dummies.to_csv(data_set_path.replace('.csv', '_xgbdummies.csv'))
    else:
        data_w_dummies.to_csv(data_set_path.replace('current.csv', 'xgbdummies_current.csv'))

    dummies_only = data_w_dummies[new_cols]
    save_path = data_set_path.replace('.csv', '_dummies_groupscore.csv')
    print('Saved dummy variable (only) dataset ' + save_path)
    dummies_only.to_csv(save_path)

    raw_data.drop(cat_cols,axis=1,inplace=True)
    save_path = data_set_path.replace('.csv', '_nocat.csv')
    print('Saved no category dataset ' + save_path)
    raw_data.to_csv(save_path)



import pandas as pd
import matplotlib.pyplot as plt
from math import ceil

def event_count_plot(qa_data_path, event_name,**kwargs):
    event_data_path = qa_data_path + '_' + event_name + '.csv'
    qa_data_df=pd.read_csv(event_data_path)
    plt.figure(figsize=(6, 4))
    plt.plot('event_date', 'n_event', data=qa_data_df, marker='', color='black', linewidth=2)
    plt.ylim(0, ceil(1.1 * qa_data_df['n_event'].dropna().max()))
    plt.title('{} event count'.format(event_name))
    plt.gca().figure.autofmt_xdate()
    plt.xticks(list(filter(lambda x:x.endswith(("01")),qa_data_df['event_date'].tolist())))
    plt.tight_layout()
    save_to_path=event_data_path.replace('.csv', '_' + event_name + '_event_qa.png')
    print('Saving metric qa plot to ' + save_to_path)
    plt.savefig(save_to_path)
    plt.close()





import pandas as pd
import numpy as np
import os
from shutil import copyfile

from customer import Customer

def is_pos_def(x):
    '''
    Quick check of whether the given matrix is positive definite
    https://stackoverflow.com/questions/16266720/find-out-if-matrix-is-positive-definite-with-numpy
    https://stackoverflow.com/questions/42908334/checking-if-a-matrix-is-symmetric-in-numpy
    :param x:
    :return: true/false
    '''
    rtol = 1e-05
    atol = 1e-08
    return np.all(np.linalg.eigvals(x) > 0) and np.allclose(x, x.T, rtol=rtol, atol=atol)

class BehaviorModel:

    def generate_customer(self):
        '''
        This is the main result expected of the behaviora model: produce customers that can be simulated.
        Implemented by sub-classes, so this definition serves as an interface definition.
        :return:
        '''
        raise NotImplementedError('Sub-classes must define generate_customer!')

    def insert_event_types(self,schema_name,db):
        '''
        A second function of the behavior model is to make sure the database is ready to accept the type of events
        that it produces. This does that by scanning the event type table and inserting any events produced by the model
        that are not alraedy in the table.   Insert the event types for a given schema into the database.
        :param schema_name: string name of the schenma
        :param db: postgres.Posrtgres
        :return:
        '''
        for idx,e in enumerate(self.behave_names):
            id = db.one(self.eventIdSql(schema_name, e))
            if id is None:
                db.run("INSERT into %s.event_type VALUES (%d,'%s');" % (schema_name,idx,e) )

    def eventIdSql(self,schema, event_name):
        '''
        SQL to check if an event id is already in the event type table
        :param schema:
        :param event_name:
        :return:
        '''
        return "select event_type_id from %s.event_type where event_type_name='%s'" % (schema, event_name)

class GaussianBehaviorModel(BehaviorModel):

    def __init__(self,name,random_seed=None,version='model'):
        '''
        This behavior model uses a mean and (pseudo) covariance matrix to generate customers with event rates.
        The parameters are passed on a csv file that should be located in a `conf` directory adjacent to the code.
        The format of the data file is:
            first column of the data file should be the names and must have the heading 'behavior'
            second colum is the mean rates and must have the heading 'mean'
            the remaining columns should be a pseudo-covariance matrix for the behaviors, so it must be real valued and
            have the right number of columns with column names the same as the rows
        These are loaded using pandas.
        :param name:
        '''
        self.name=name
        self.version=version
        model_path='../conf/'+name + '_' + version + '.csv'
        model=pd.read_csv(model_path)
        model.set_index(['behavior'],inplace=True)
        self.behave_means=model['mean']
        self.behave_names=model.index.values
        self.behave_cov=model[self.behave_names]
        self.min_rate=0.01*self.behave_means.min()
        corr_scale=(np.absolute(self.behave_cov.to_numpy()) <= 1.0).all()
        if random_seed is not None:
            np.random.seed(random_seed)
        if not is_pos_def(self.behave_cov):
            if input("Matrix is not positive semi-definite: Multiply by transpose? (enter Y to proceed)") in ('y','Y'):
                # https://stackoverflow.com/questions/619335/a-simple-algorithm-for-generating-positive-semidefinite-matrices
                self.behave_cov= np.dot(self.behave_cov, self.behave_cov.transpose())
            else:
                exit(0)

        if corr_scale:
            self.scale_correlation_to_covariance()

        # Save to a csv
        save_path = '../../../fight-churn-output/' + name + '/'
        os.makedirs(save_path, exist_ok=True)
        copy_path = save_path + name + '_simulation_model.csv'
        copyfile(model_path, copy_path)

        # For debugging
        # np.savetxt('../conf/'+name+ '_behavior_cov.csv', self.behave_cov,delimiter=',')
        # std_ = np.sqrt(np.diag(self.behave_cov))
        # corr = self.behave_cov / np.outer(std_, std_)
        # np.savetxt('../conf/'+name+ '_behavior_corr.csv', corr,delimiter=',')
        # exit(0)


    def scale_correlation_to_covariance(self):
        # print('Scaling correlation by behavior means...')
        # This seems to give a reasonable amount of variance if the matrix was designed as a set of correlations
        scaling = np.sqrt(self.behave_means.abs() * np.sqrt(self.behave_means.abs()))
        self.behave_cov = np.matmul(self.behave_cov, np.diag(scaling))
        self.behave_cov = np.matmul(np.diag(scaling), self.behave_cov)

    def behave_var(self):
        return np.diagonal(self.behave_cov)

    def generate_customer(self):
        '''
        Given a mean and covariance matrix, the event rates for the customer are drawn from the multi-variate
        gaussian distribution.
        :return: a Custoemr object
        '''
        customer_rates=np.random.multivariate_normal(mean=self.behave_means,cov=self.behave_cov)
        customer_rates=customer_rates.clip(min=self.min_rate) # clip : no negative rates!
        new_customer= Customer(customer_rates)
        # print(customer_rates)
        return new_customer


class FatTailledBehaviorModel(GaussianBehaviorModel):

    def __init__(self,name,random_seed=None,version=None):
        self.exp_base = 1.6
        self.log_fun = lambda x: np.log(x) / np.log(self.exp_base)
        self.exp_fun = lambda x: np.power(self.exp_base,x)

        super(FatTailledBehaviorModel,self).__init__(name,random_seed,version)

    def scale_correlation_to_covariance(self):
        self.log_means=self.log_fun(self.behave_means)
        rectified_means =np.array([max(m,0.0) for m in self.log_means])
        # print('Scaling correlation by behavior means...')

        scaling = np.sqrt(rectified_means)
        self.behave_cov = np.matmul(self.behave_cov, np.diag(scaling))
        self.behave_cov = np.matmul(np.diag(scaling), self.behave_cov)

    def behave_var(self):
        return self.exp_fun( np.diagonal(self.behave_cov))

    def generate_customer(self,start_of_month):
        '''
        Given a mean and covariance matrix, the event rates for the customer are drawn from the multi-variate
        gaussian distribution.
        subtract 0.5 and set min at 0.5 per month, so there can be very low rates despite 0 (1) min in log normal sim
        :return: a Custoemr object
        '''
        customer_rates=np.random.multivariate_normal(mean=self.log_means,cov=self.behave_cov)
        customer_rates=self.exp_fun(customer_rates)
        customer_rates = np.maximum(customer_rates-0.667,0.333)
        new_customer= Customer(customer_rates,channel_name=self.version,start_of_month=start_of_month)
        # print(customer_rates)
        return new_customer



import pandas as pd
import numpy as np
import os
from listing_7_5_fat_tail_scores import transform_fattail_columns, transform_skew_columns

def rescore_metrics(data_set_path):

    load_mat_df = reload_churn_data(data_set_path,'load_mat','6.4',is_customer_data=False)
    score_df = reload_churn_data(data_set_path,'score_params','7.5',is_customer_data=False)
    current_data = reload_churn_data(data_set_path,'current','8.3',is_customer_data=True)
    assert set(score_df.index.values)==set(current_data.columns.values),"Data to re-score does not match transform params"
    assert set(load_mat_df.index.values)==set(current_data.columns.values),"Data to re-score does not match lodaasasdfasdfasdf matrix"

    transform_skew_columns(current_data,score_df[score_df['skew_score']].index.values)
    transform_fattail_columns(current_data,score_df[score_df['fattail_score']].index.values)
    scaled_data = score_current_data(current_data,score_df,data_set_path)
    grouped_data = group_current_data(scaled_data, load_mat_df,data_set_path)
    save_segment_data(grouped_data,current_data,load_mat_df,data_set_path)

def score_current_data(current_data,score_df, data_set_path):
    current_data=current_data[score_df.index.values]
    scaled_data=(current_data-score_df['mean'])/score_df['std']
    score_save_path=data_set_path.replace('.csv','_current_scores.csv')
    scaled_data.to_csv(score_save_path,header=True)
    print('Saving score results to %s' % score_save_path)
    return scaled_data

def group_current_data(scaled_data,load_mat_df,data_set_path):
    scaled_data = scaled_data[load_mat_df.index.values]
    grouped_ndarray = np.matmul(scaled_data.to_numpy(), load_mat_df.to_numpy())
    current_data_grouped = pd.DataFrame(grouped_ndarray,columns=load_mat_df.columns.values, index=scaled_data.index)
    score_save_path=data_set_path.replace('.csv','_current_groupscore.csv')
    current_data_grouped.to_csv(score_save_path,header=True)
    print('Saving grouped results to %s' % score_save_path)
    return current_data_grouped

def save_segment_data(current_data_grouped, current_data, load_mat_df, data_set_path):
    group_cols =  load_mat_df.columns[load_mat_df.astype(bool).sum(axis=0) > 1]
    no_group_cols = load_mat_df.columns[load_mat_df.astype(bool).sum(axis=0) == 1]
    segment_df = current_data_grouped[group_cols].join(current_data[no_group_cols])
    segment_df.to_csv(data_set_path.replace('.csv','_current_groupmets_segment.csv'),header=True)

def reload_churn_data(data_set_path,suffix,listing,is_customer_data):
    data_path = data_set_path.replace('.csv', '_{}.csv'.format(suffix))
    assert os.path.isfile(data_path),'You must run listing {} to save {} first'.format(listing,suffix)
    ic = [0,1] if is_customer_data else 0
    churn_data = pd.read_csv(data_path, index_col=ic)
    return churn_data


import pandas as pd
import matplotlib.pyplot as plt

def metric_pair_plot(data_set_path, metric1='',metric2=''):

    churn_data = pd.read_csv(data_set_path,index_col=[0,1])

    met1_series = churn_data[metric1]
    met2_series = churn_data[metric2]

    corr = met1_series.corr(met2_series)

    plt.scatter(met1_series, met2_series, marker='.')

    plt.xlabel(metric1)
    plt.ylabel(metric2)
    plt.title('Correlation = %.2f' % corr)
    plt.tight_layout()
    plt.grid()

    save_name = data_set_path.replace('.csv', '_' + metric1 + '_vs_' + metric2 + '.png')
    plt.savefig(save_name )
    print('Saving plot to %s' % save_name)
    plt.close()
